/*
 * tkresolver - Telekinesis testing software
 *
 * Copyright (C) 2004 Robert Ham <rah@bash.sh>
 *
 * GPL all the way, baby :)
 *
 */

#include <time.h>

#include <sstream>

#include "libtkcommon/mbs.h"
#include "libtkcommon/exceptions.h"
#include "libtkcommon/debug.h"
#include "database.h"

using namespace TkResolver;
using namespace Db;

static const char *
get_month_str(unsigned month)
{
  switch (month)
    {
    case  1: return "JAN";
    case  2: return "FEB";
    case  3: return "MAR";
    case  4: return "APR";
    case  5: return "MAY";
    case  6: return "JUN";
    case  7: return "JUL";
    case  8: return "AUG";
    case  9: return "SEP";
    case 10: return "OCT";
    case 11: return "NOV";
    case 12: return "DEC";
    }

  throw std::domain_error("unknown month number");
}

String
TkResolver::Db:: date_time_string(time_t t)
{
  struct tm *ltm;
  
  ltm = localtime (&t);
  if (!ltm)
    {
      throw std::runtime_error("could not convert time to local time components");
    }
  ltm->tm_year += 1900;
  ltm->tm_mon  += 1;

  std::ostringstream strm;
  strm << ltm->tm_year
       << "-"
       << (ltm->tm_mon < 10 ? "0" : "")
       << ltm->tm_mon
       << "-"
       << (ltm->tm_mday < 10 ? "0" : "")
       << ltm->tm_mday
       << " "
       << ltm->tm_hour
       << ":"
       << ltm->tm_min
       << ":"
       << ltm->tm_sec;

  return strm.str();
}


Error:: Error (mysqlpp::Connection &connection)
  : std::runtime_error(connection.error())
{
}


Database:: Database (const TkString &hostname,
		     const TkString &database,
		     const TkString &username,
		     const TkString &password)
  : _connection(std::string(database).c_str(),
		std::string(hostname).c_str(),
		std::string(username).c_str(),
		std::string(password).c_str())
{
  if (!_connection.connected())
    throw Error(_connection);

  create_table_info_query();
}

Database:: ~Database()
{
  delete _table_info_query;
}

void
Database:: create_table_info_query()
{
  _table_info_query = new mysqlpp::Query(&_connection);
  (*_table_info_query) << mysqlpp::SQLString("select * from %0:table t");
  _table_info_query->parse();
}

void
Database:: get_table_info(const Name &name)
{
  mysqlpp::ResUse result = _table_info_query->use(mysqlpp::SQLString(name));
  if (!result)
    throw Error(_connection);

  const mysqlpp::FieldNames names (&result);
  const mysqlpp::FieldTypes types (&result);

  FieldInfo fields;
  unsigned  i = 0;
  for (mysqlpp::FieldNames::const_iterator name_iter = names.begin();
       name_iter != names.end();
       ++name_iter, ++i)
    {
      
      fields.insert(FieldInfo::value_type(*name_iter, types[i]));
    }

  _table_info.insert(TableInfo::value_type(name, fields));

  result.purge();
}

void
Database:: check_table_info(const Name &name)
{
  if (_table_info.find(name) == _table_info.end())
    get_table_info(name);
}

Type
Database:: get_column_type(const Name &table_name, const Name &column_name)
{
  check_table_info(table_name);

  const FieldInfo &fields(_table_info[table_name]);

  FieldInfo::const_iterator iter = fields.find(column_name);
  if (iter == fields.end())
    throw std::runtime_error(std::string("could not find column ") + table_name + "." + column_name);
  
  return iter->second;
}

DataSet *
Database:: get_table_data(const Name &table_name, const NameList &column_names)
{
  mysqlpp::Query query(&_connection);

  query << "select ";
  NameList::const_iterator iter, next;
  for (iter = column_names.begin();
       iter != column_names.end();
       iter = next)
    {
      next = iter;
      ++next;

      query << *iter;
      if (next != column_names.end())
	query << ",";
    }
  query << " from "
	<< table_name
	<< " t";

  DataSet * result = new DataSet(query.store());

  if (!(*result))
    throw Error(_connection);

  return result;
}

Id
Database:: insert_table_data(const Name &table_name, const Parameters &parameters)
{
  mysqlpp::Query query(&_connection);
  std::ostringstream values;
  mysqlpp::SQLQueryParms parms;

  query << "insert into " << table_name << "(";
  values << ") values (";

  Parameters::const_iterator iter, next;
  unsigned i = 0;
  for(iter = parameters.begin();
      iter != parameters.end();
      iter = next)
    {
      next = iter; ++next;

      
      query << iter->first;
	  
      if (iter->second)
	{
	  values << "%" << i++ << "r:" << iter->first;
	  parms << iter->second;
	}
      else
	values << "NULL";

      if(next != parameters.end())
	{
	  query << ", ";
	  values << ", ";
	}
	  
    }
  query << values.str();
  query << ")";
  
  /*  WDEBUG_OUT << TkString(query.str()) << std::endl;
  WDEBUG_OUT << L"parms count: " << parms.size() << std::endl;
  for (mysqlpp::SQLQueryParms::iterator iter( parms.begin() );
       iter != parms.end();
       ++iter)
    {
      try
	{
	  WDEBUG_OUT << L"parm: " << TkString(*iter) << std::endl;
	}
      catch (...) { WDEBUG_OUT << L"parm skipped" << std::endl; }
      } */
  
  query.parse();
  

  mysqlpp::ResNSel result(query.execute(parms));
  if (!result)
    throw Error(_connection);

  return result.insert_id;
}

void
Database:: get_experiment_group_parameters(const Id &experiment_group,
					   TkString * algorithm_spec,
					   unsigned * bit_desc,
					   unsigned short * traces,
					   unsigned * total_time,
					   double * speed,
					   double * horizon)
{
  mysqlpp::Query query(&_connection);

  query << " select algorithm_spec, bit_descriptor, traces, time, speed, horizon";
  query << " from experiment_group where experiment_group_id = %0:experiment_group";
  query.parse();
  
  DataSet result(query.store(experiment_group));

  if (!result)
    throw Error(_connection);

  //  DEBUG_OUT << "experiment group parameters row count: " << result.num_rows() << std::endl;

  unsigned field = 0;
  if (algorithm_spec)
    {
      //      DEBUG_OUT << "getting parameter algorithm spec" << std::endl;
      std::string c_algorithm_spec( result.at(0).at(field++) );
      //      DEBUG_OUT << "got parameter" << std::endl;
      *algorithm_spec = c_algorithm_spec;
    }

  if (bit_desc)
    {
      //      DEBUG_OUT << "getting parameter bit desc" << std::endl;
    *bit_desc   = result.at(0).at(field++);
    //      DEBUG_OUT << "got parameter" << std::endl;
    }
  if (traces)
    {
      //      DEBUG_OUT << "getting parameter traces" << std::endl;
    *traces     = result.at(0).at(field++);
    //      DEBUG_OUT << "got parameter" << std::endl;
    }
  if (total_time)
    {
      //      DEBUG_OUT << "getting parameter total time" << std::endl;
    *total_time = result.at(0).at(field++);
    //      DEBUG_OUT << "got parameter" << std::endl;
    }
  if (speed)
    {
      //      DEBUG_OUT << "getting parameter speed" << std::endl;
    *speed      = result.at(0).at(field++);
    //      DEBUG_OUT << "got parameter" << std::endl;
    }
  if (horizon && speed && *speed > 0.0)
    {
      //      DEBUG_OUT << "getting parameter horizon" << std::endl;
    *horizon    = result.at(0).at(field++);
    //      DEBUG_OUT << "got parameter" << std::endl;
    }
}

Id
Database:: insert_experiment(const Id &experiment_group,
			     const Db::CodeValue &intent_option_code,
			     const time_t &date_time,
			     const TkString &location_name,
			     const TkString &notes,
			     const TkDouble &ability_rating,
			     unsigned width,
			     unsigned height,
			     const TkString &retriever,
			     bool retriever_is_true_random,
			     bool retriever_specified,
			     const Random::Bits &randomness)
{
  Parameters params;

  /*  StringValue str;
  
  str = experiment_group;                     params["experiment_group_id"] = str;
  str = date_time_string(date_time);          params["date_time"] = str;
  str = static_cast<String>(location_name);   params["location_name"] = str;
  str = static_cast<String>(notes);           params["notes"] = str;
  str = ability_rating;                       params["ability_rating"] = str;
  str = width;                                params["width"] = str;
  str = height;                               params["height"] = str;
  str = static_cast<String>(retriever);       params["retriever"] = str;
  str = retriever_is_true_random;             params["retriever_is_true_random"] = str;
  str = retriever_specified;                  params["retriever_specified"] = str;
  str = randomness.get_data();                params["randomness"] = str;
  str = randomness.size();                    params["randomness_bits"] = str;
  str = intent_option_code;                   params["intent_option_code"] = str; */

  params["experiment_group_id"]         = experiment_group;
  params["date_time"]                   = date_time_string(date_time);
  params["location_name"]               = static_cast<String>(location_name);
  params["notes"]                       = static_cast<String>(notes);
  params["ability_rating"]              = ability_rating;
  params["width"]                       = width;
  params["height"]                      = height;
  params["retriever"]                   = static_cast<String>(retriever);
  params["retriever_is_true_random"]    = retriever_is_true_random;
  params["retriever_specified"]         = retriever_specified;
  params["randomness"]                  = randomness.get_data();
  params["randomness_bits"]             = randomness.size();
  params["intent_option_code"]          = intent_option_code;
  
  return insert_table_data("experiment", params);
}

void
Database:: insert_experiment(const Id &experiment_id,
			     const IdList &operators)
{
  for (IdList::const_iterator iter(operators.begin());
       iter != operators.end();
       ++iter)
    {
      Parameters params;
      
      params["experiment_id"]  = experiment_id;
      params["operator_id"] = *iter;
      
      insert_table_data("experiment_operator", params);
    }
}

void
Database:: insert_experiment(const Id &experiment_id,
			     const Result &result)
{
  unsigned ntraces( result.begin()->size() );
  unsigned set_index(0);
  unsigned offset_index;
  for (Result::const_iterator iter(result.begin());
       iter != result.end();
       ++iter, ++set_index)
    {
      Parameters result_params;
      result_params["experiment_id"] = experiment_id;
      result_params["set_index"]     = set_index;
      insert_table_data("experiment_result", result_params);


      for (offset_index = 0; offset_index < ntraces; ++offset_index)
	{
	  Parameters offset_params;
	  offset_params["experiment_id"] = experiment_id;
	  offset_params["set_index"]     = set_index;
	  offset_params["offset_index"]  = offset_index;
	  offset_params["offset"]        = (*iter)[offset_index];
	  insert_table_data("experiment_result_offset", offset_params);
	}
    }
}

Id
Database:: insert_experiment(const Id &experiment_group,
			     const Db::CodeValue &intent_option_code,
			     const time_t &date_time,
			     const TkString &location_name,
			     const TkString &notes,
			     const TkDouble &ability_rating,
			     unsigned width,
			     unsigned height,
			     const TkString &retriever,
			     bool retriever_is_true_random,
			     bool retriever_specified,
			     const Random::Bits &randomness,
			     const IdList &operators,
			     const Result &result)
{
  Id id = insert_experiment(experiment_group,
			    intent_option_code,
			    date_time,
			    location_name,
			    notes,
			    ability_rating,
			    width,
			    height,
			    retriever,
			    retriever_is_true_random,
			    retriever_specified,
			    randomness);

  insert_experiment(id, operators);
  insert_experiment(id, result);
  
  return id;
}


Db::Id
Database:: lookup_experiment_group(const Name &name)
{
  mysqlpp::Query query(&_connection);

  query << " select experiment_group_id";
  query << " from experiment_group where name = %0r:name";
  query.parse();
  
  DataSet result(query.store(String(name)));

  if (!result)
    throw Error(_connection);
  
  return result.at(0).at(0);
}

Id
Database:: choose_experiment_group()
{
  mysqlpp::Query query(&_connection);

  query << " select   eg.experiment_group_id";
  query << " from     experiment_group eg";
  query << " ,        experiment e";
  query << " where    e.experiment_group_id = eg.experiment_group_id";
  query << " order by e.date_time desc";
  query.parse();
  
  DataSet result(query.store());

  if (!result)
    {
      return 1;
      throw Error(_connection);
    }
  
  try {
    return result.at(0).at(0);
  } catch (...) {
    return 1;
  }
}

TkString
Database:: choose_location()
{
  mysqlpp::Query query(&_connection);

  //  select l.*, e.date_time from location l, experiment e where e.location_name = l.name order by e.date_time desc;
  query << " select   l.name";
  query << " from     location l";
  query << " ,        experiment e ";
  query << " where    e.location_name = l.name";
  query << " order by e.date_time desc";
  query.parse();
  
  DataSet result(query.store());

  if (!result)
    {
      return "44 Greville Road";
      throw Error(_connection);
    }
  
  try {
    return static_cast<const char *>(result.at(0).at(0));
  } catch (...) {
    return "44 Greville Road";
  }
}

Id
Database:: choose_operator()
{
  mysqlpp::Query query(&_connection);

  query << " select   o.operator_id";
  query << " from     operator o";
  query << " ,        experiment e";
  query << " ,        experiment_operator eo";
  query << " where    eo.operator_id = o.operator_id";
  query << " and      e.experiment_id = eo.experiment_id";
  query << " order by e.date_time desc";
  query.parse();
  
  DataSet result(query.store());

  if (!result)
    {
      return 1;
      throw Error(_connection);
    }

  try {
    return result.at(0).at(0);
  } catch (...) {
    return 1;
  }
}

Code
Database:: choose_intent_option(const Id &experiment_group)
{
  mysqlpp::Query query(&_connection);

  query << " select   eio.code";
  query << " from     experiment_group eg";
  query << " ,        experiment_intent_option eio";
  query << " where    eg.experiment_group_id = %0:group_id";
  query << " and      eio.experiment_intent_id = eg.experiment_intent_id";
  query << " order by (";
  query << "             select   count(*)";
  query << "             from     experiment";
  query << "             where    experiment_group_id = eg.experiment_group_id";
  query << "             and      intent_option_code = eio.code";
  query << "	      )";

  query.parse();
  
  DataSet result(query.store(String(experiment_group)));

  if (!result)
    {
      throw Error(_connection);
    }
  
  try {
    return Code(result.at(0).at(0));
  } catch (...) {
    return "BL";
  }
}

Codes
Database:: get_intent_options(const Id &experiment_group)
{
  mysqlpp::Query query(&_connection);
  
  query << " select   distinct e.intent_option_code";
  query << " from     experiment e";
  query << " where    e.experiment_group_id   = %0:group_id";
  
  query.parse();
  
  DataSet result(query.store(String(experiment_group)));
  
  if (!result)
    throw Error(_connection);
  
  Codes codes;
  unsigned rows = result.rows();
  for (unsigned i = 0; i < rows; ++i)
    codes.push_back(Code(result.at(i).at(0)));
       
  return codes;
}

IdList
Database:: get_experiments(const Id &experiment_group, const Code &intent_option_code)
{
  mysqlpp::Query query(&_connection);
  query << " select   experiment_id"
        << " from     experiment"
	<< " where    experiment_group_id           =  %0r:group_id";

  if (intent_option_code.size() > 0 && intent_option_code != "NULL")
    query << " and    intent_option_code =  %1r:intent_option_code";
  else
    query << " and    intent_option_code is null";
 
  query.parse();

  query.def["group_id"]  = String(experiment_group);
  
  if (intent_option_code.size() > 0 && intent_option_code != "NULL")
    query.def["intent_option_code"] = String(intent_option_code);

  DataSet result(query.store());
  
  if (!result)
    throw Error(_connection);
  
  IdList experiments;
  unsigned long rows = result.rows();
  for (unsigned long i = 0; i < rows; ++i)
    experiments.push_back(Id(result.at(i).at(0)));
       
  return experiments;
}

DataCursor *
Database:: get_results(const Id &experiment)
{
  mysqlpp::Query query(&_connection);
  query << " select   offset"
        << " from     experiment_result_offset"
	<< " where    experiment_id =  %0:experiment";

  query.parse();

  //  query.def["experiment_id"]  = experiment;

  DataCursor * result = new DataCursor(query.use(String(experiment)));
  if (!(*result))
    throw Error(_connection);

  return result;
}

unsigned
Database:: get_experiment_screen_width(const Id &experiment)
{
  unsigned width;
  mysqlpp::Query query(&_connection);

  query << " select width";
  query << " from experiment where experiment_id = %0:experiment;";
  query.parse();

  DataSet result(query.store(experiment));

  if (!result)
    throw Error(_connection);

  return result.at(0).at(0);
}
